/*----------------------------------------------------------------------
  File    : lvq.c
  Contents: learning vector quantization with some extensions
  Author  : Christian Borgelt
  History : 2003.02.17 file created from file cluster.c
            2003.03.09 first compilable version completed
            2003.03.11 weight parameter added to function lvq_reg
            2003.03.15 organization of quantization prototypes changed
            2003.05.15 activation normalization modes added
            2003.08.06 bug in mapping binary attributes fixed
            2003.08.11 adapted to new module attmap
            2003.08.12 adapted to new module nstats
            2003.08.15 adapted to new module radfn
            2007.01.10 use of attribute maps changed
            2013.08.19 adapted to definition of type DIMID in matrix.h
            2018.09.10 extended functions (attset, attmap) removed
----------------------------------------------------------------------*/
#include <stdlib.h>
#include <limits.h>
#include <float.h>
#include <math.h>
#include <string.h>
#include <assert.h>
#include "lvq.h"
#ifdef STORAGE
#include "storage.h"
#endif

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
#define BLKSIZE      16         /* block size for prototype vector */
#define MINVAR       1e-12      /* minimal variance */

/*----------------------------------------------------------------------
  Constants
----------------------------------------------------------------------*/
static char *nrmnames[] =       /* names of normalization modes */
  { "none", "sum1", "max1", "hard" };

/*----------------------------------------------------------------------
  Functions
----------------------------------------------------------------------*/

static LVQNET* create (DIMID dim, DIMID cnt)
{                               /* --- create an LVQ network */
  LVQNET *lvq;                  /* created LVQ network */

  assert((dim > 0) && (cnt >= 0)); /* check the function arguments */
  lvq = (LVQNET*)malloc(sizeof(LVQNET));
  if (!lvq) return NULL;        /* create the base structure */
  lvq->dim       = dim;         /* and initialize some fields */
  lvq->cnt       = cnt;
  lvq->init      = 0;
  lvq->reg       = lvq->spw = 0;
  lvq->type      = LVQ_CENTER;  /* prototypes consist of centers only */
  lvq->actfn     = rf_cauchy;   /* default activation function: */
  lvq->params[0] = 2;           /* act(x) = 1/d^2(x,c) */
  lvq->params[1] = 0;           /* (inverse sq. Euclidean distance) */
  lvq->mode      = LVQ_HARD;
  lvq->wtarf     = lvq->exp   = 0;
  lvq->lrcur     = lvq->lrate = 0.01;
  lvq->decay     = -0.01;
  lvq->szscl     = 1;
  lvq->steps     = 1;
  lvq->qps       = NULL;        /* initialize pointers, */
  lvq->vec       = NULL;        /* so that deletion */
  lvq->nst       = NULL;        /* with lvq_delete() works */
  return lvq;                   /* return the created LVQ network */
}  /* create() */

/*--------------------------------------------------------------------*/

static int qprot (QPROT *p, DIMID dim)
{                               /* --- init. a quant. prototype */
  assert(p && (dim > 0));       /* check the function arguments */
  p->ctr = (double*)malloc(3 *(size_t)dim *sizeof(double));
  if (!p->ctr) return -1;       /* create and */
  p->dif = p->ctr +dim;         /* organize the vectors, */
  p->chg = p->dif +dim;         /* and initialize the fields */
  p->var = 1;
  p->sqr = p->sum = p->spw = 0;
  return 0;                     /* return 'ok' */
}  /* qprot() */

/*--------------------------------------------------------------------*/

LVQNET* lvq_create (DIMID dim, DIMID cnt)
{                               /* --- create a LVQ network */
  DIMID  i;                     /* loop variable */
  LVQNET *lvq;                  /* created neural network */
  QPROT  *p;                    /* to traverse the prototypes */

  assert((dim > 0) && (cnt > 0));  /* check the function arguments */
  lvq = create(dim, cnt);       /* create the base structure */
  if (!lvq) return NULL;        /* and initialize some fields */
  lvq->qps = p = (QPROT*)malloc((size_t)cnt *sizeof(QPROT));
  if (!lvq->qps) { lvq_delete(lvq); return NULL; }
  for (p += i = lvq->cnt; --i >= 0; )
    (--p)->ctr = NULL;          /* clear pointers for cleanup */
  for (p += i = lvq->cnt; --i >= 0; )
    if (qprot(--p, dim) != 0) { /* init. the quantization prototypes */
      lvq_delete(lvq); return NULL; }
  lvq->vec = (double*)malloc(2 *(size_t)dim *sizeof(double));
  if (!lvq->vec) { lvq_delete(lvq); return NULL; }
  lvq->buf = lvq->vec +dim;     /* create the numeric vectors */
  lvq->nst = nst_create(dim);   /* and the numerical statistics */
  if (!lvq->nst) { lvq_delete(lvq); return NULL; }
  return lvq;                   /* return the created neural network */
}  /* lvq_create() */

/*--------------------------------------------------------------------*/

void lvq_delete (LVQNET *lvq)
{                               /* --- delete a LVQ network */
  DIMID i;                      /* loop variable */

  assert(lvq);                  /* check the function argument */
  if (lvq->nst) nst_delete(lvq->nst);
  if (lvq->vec) free(lvq->vec); /* delete the numeric vectors */
  if (lvq->qps) {               /* if there is a prototypes vector */
    for (i = 0; i < lvq->cnt; i++)
      if (lvq->qps[i].ctr) free(lvq->qps[i].ctr);
    free(lvq->qps);             /* delete all centers and */
  }                             /* the prototype vector */
  free(lvq);                    /* delete the base structure */
}  /* lvq_delete() */

/*--------------------------------------------------------------------*/

void lvq_value (LVQNET *lvq, DIMID index, double value)
{                               /* --- set an input value */
  assert(lvq                    /* check the function arguments */
  &&    (index >= 0) && (index < lvq->dim));
  lvq->vec[index] = (value -nst_offset(lvq->nst, index))
                           *nst_factor(lvq->nst, index);
}  /* lvq_value() */

/*--------------------------------------------------------------------*/

void lvq_type (LVQNET *lvq, int type, double radius)
{                               /* --- set the cluster type */
  DIMID i;                      /* loop variable */

  assert(lvq);                  /* check the function arguments */
  lvq->type = type;             /* note the prototype type */
  if (radius <= 0) radius = 1;  /* check and adapt the initial radius */
  radius *= radius;             /* and compute the variance */
  for (i = lvq->cnt; --i >= 0; )
    lvq->qps[i].var = radius;   /* note the initial variance */
}  /* lvq_type() */

/*--------------------------------------------------------------------*/

void lvq_init (LVQNET *lvq, int mode, double range,
               RANDFN randfn, double *vec)
{                               /* --- initialize a LVQ network */
  DIMID  i, k, n;               /* loop variables */
  QPROT  *p;                    /* to traverse the prototypes */
  NSTATS *nst;                  /* simple numerical statistics */
  double *c, *b;                /* center vector, buffer */
  double x, d, m;               /* coordinate buffers */

  assert(lvq && randfn);        /* check the function arguments */

  /* --- initialize other parameters --- */
  if ((mode != LVQ_POINTS)      /* if not called a second time */
  ||  (lvq->init <= 0) || (lvq->init >= lvq->cnt)) {
    lvq->steps = 1;             /* init. the update step counter */
    for (p = lvq->qps +(k = lvq->cnt); --k >= 0; ) {
      (--p)->var = 1.0;         /* traverse the prototypes and */
      p->sqr = p->sum = p->spw = 0.0;  /* init. the parameters */
      memset(p->chg, 0, (size_t)lvq->dim *sizeof(double));
    }                           /* clear the coordinates changes */
  }

  /* --- initialize reference vectors --- */
  nst = lvq->nst;               /* get the numerical statistics */
  switch (mode) {               /* evaluate the initialization mode */

    case LVQ_POINTS:            /* -- given points in the data space */
      if (!vec) vec = lvq->vec;
      if (lvq->init >= lvq->cnt)
        lvq->init = 0;          /* if all protos. are init., restart */
      p = lvq->qps +lvq->init++;
      nst_norm(nst, vec, p->ctr);
      if (range <= 0) return;   /* copy and scale a given data point */
      for (c = p->ctr, i = lvq->dim; --i >= 0; )
        c[i] += (2 *randfn() -1) *range;
      return;                   /* and add a random offset */

    case LVQ_CENTER:            /* -- center of the data space */
      c = lvq->qps[0].ctr;      /* compute the center */
      nst_center(nst, c);       /* of the data space */
      for (k = lvq->cnt; --k > 0; )
        vec_copy(lvq->qps[k].ctr, c, lvq->dim);
      break;                    /* copy the center to all prototypes */

    case LVQ_DIAG:              /* -- diagonal of the data space */
    case LVQ_LATIN:             /* -- latin hypercube sampling */
      for (i = lvq->dim; --i >= 0; ) {
        m = nst_max(nst, i);    /* compute value decrement */
        d = (m -nst_min(nst, i)) / (double)lvq->cnt;
        x = m -0.5*d;           /* compute last value */
        for (k = lvq->cnt; --k >= 0; ) {
          lvq->qps[k].ctr[i] = x; x -= d; }
      }                         /* set equally spaced values */
      if (mode == LVQ_DIAG)     /* if only to set the diagonal, */
        break;                  /* there is nothing else to be done */

    /* case LVQ_LATIN: */       /* -- latin hypercube sampling */
      p = lvq->qps;             /* shuffle elements of the vectors */
      for (n = lvq->cnt; --n > 0; ) {
        for (i = lvq->dim; --i >= 0; ) {
          k = (DIMID)((double)(n+1) *randfn());
          if      (k > n) k = n;   /* compute a random index in */
          else if (k < 0) k = 0;   /* the remaining set of vectors */
          x           = p[k].ctr[i];
          p[k].ctr[i] = p[n].ctr[i];
          p[n].ctr[i] = x;      /* exchange the i-th elements of */
        }                       /* the k-th and the n-th center */
      } break;                  /* (shuffle dimensions independently) */

    case LVQ_UNIFORM:           /* -- uniformly distributed */
    default:                    /* (this is also the default) */
      nst_spans(nst, b = lvq->buf);    /* get the value spans */
      for (p = lvq->qps +(k = lvq->cnt); --k >= 0; )
        for (c = (--p)->ctr, i = lvq->dim; --i >= 0; )
          c[i] =  nst_min(nst, i) +b[i] *randfn();
      break;                    /* set reference vectors to */
  }                             /* random points in the data space */

  /* --- scale the coordinates --- */
  for (p = lvq->qps +(k = lvq->cnt); --k >= 0; ) {
    c = (--p)->ctr; nst_norm(nst, c, c); }

  /* --- add a random offset --- */
  if (range > 0) {              /* if a range for offsets is given */
    for (p = lvq->qps +(k = lvq->cnt); --k >= 0; )
      for (c = (--p)->ctr, i = lvq->dim; --i >= 0; )
        c[i] += (2 *randfn() -1) *range;
  }                             /* add a random offset to all values */
}  /* lvq_init() */

/*--------------------------------------------------------------------*/

void lvq_actfn (LVQNET *lvq, RADFN actfn, double *params)
{                               /* --- set the activation function */
  assert(lvq && actfn && params);  /* check the function arguments */
  lvq->actfn     = actfn;       /* note the activation function */
  lvq->params[0] = params[0];   /* and its parameters */
  lvq->params[1] = params[1];
}  /* lvq_actfn() */

/*--------------------------------------------------------------------*/

double lvq_sqrdist (LVQNET *lvq, DIMID pid, const double *vec)
{                               /* --- compute raw squared distance */
  DIMID i;                      /* loop variable */
  QPROT *p;                     /* to access the quant prototype */

  assert(lvq                    /* check the function arguments */
  &&    (pid >= 0) && (pid < lvq->cnt));
  if (vec) nst_norm(lvq->nst, vec, lvq->vec);
  vec = lvq->vec;               /* scale given vector to the buffer */
  p = lvq->qps +pid;            /* get the prototype referred to */
  for (i = 0; i < lvq->dim; i++)       /* compute difference */
    p->dif[i] = vec[i] -p->ctr[i];     /* vector to center and */
  return vec_sqrlen(p->dif, lvq->dim); /* return squared distance */
}  /* lvq_sqrdist() */

/*--------------------------------------------------------------------*/

DIMID lvq_exec (LVQNET *lvq, double *vec, double *acts)
{                               /* --- compute ouputs / activations */
  DIMID  i, k, best;            /* loop variables, prototype index */
  QPROT  *p;                    /* to traverse the prototypes */
  double *c, *b;                /* to access the center and buffer */
  double d2;                    /* squared distance of pattern */
  double sum, max;              /* sum/maximum of neuron activations */
  int    mode;                  /* activation normalization mode */

  assert(lvq);                  /* check the function arguments */
  if (vec)                      /* if a data vector is given, */
    nst_norm(lvq->nst, vec, lvq->vec); /* scale it to the buffer */
  vec = lvq->vec;               /* and then work with the buffer */

  /* --- compute absolute neuron activations --- */
  sum = 0; best = 0; max = -INFINITY; /* initialize the variables */
  for (p = lvq->qps +(k = lvq->cnt); --k >= 0; ) {
    c = (--p)->ctr;             /* traverse the prototypes */
    for (b = p->dif, i = lvq->dim; --i >= 0; )
      b[i] = vec[i] -c[i];      /* compute diff. vector to center */
    p->d2 = d2 = vec_sqrlen(b, lvq->dim);
    assert(d2 >= 0);            /* compute distance to center */
    if (lvq->type & LVQ_SIZE)   /* if to use prototype sizes, */
      d2 /= p->var;             /* divide by the variance */
    p->act = lvq->actfn(d2, lvq->params);
    assert(p->act >= 0);        /* compute the neuron output and */
    sum += p->act;              /* sum it for the normalization */
    if (p->act >= max) { max = p->act; best = k; }
  }                             /* determine the winner neuron */

  /* --- compute relative neuron activations --- */
  mode = lvq->mode;             /* get the normalization mode */
  if (lvq->wtarf > 0) {         /* if limited winner takes all */
    d2 = lvq->wtarf;            /* square the winner takes all radius */
    d2 *= d2;                   /* factor (distances are all squared) */
    if (lvq->type & LVQ_SIZE)   /* compute the squared adapted */
      d2 *= p[best].var;        /* radius for the winner neuron */
    if (p[best].d2 < d2)        /* if the data point is inside */
      mode = LVQ_HARD;          /* the adapted radius, activate */
  }                             /* according to winner takes all */
  if      (mode == LVQ_HARD) {  /* if the winner takes all */
    for (p += k = lvq->cnt; --k >= 0; )
      (--p)->act = 0;           /* set only the activation of */
    p[best].act = 1; }          /* the best neuron, clear all other */
  else if (mode == LVQ_MAX1) {  /* if to normalize to maximum 1 */
    max = (max > 0) ? 1/max :1; /* get the normalization factor */
    for (p += k = lvq->cnt; --k >= 0; )
      (--p)->act *= max; }      /* compute rel. neuron activations */
  else if (mode == LVQ_SUM1) {  /* if to normalize to sum 1 */
    sum = (sum > 0) ? 1/sum :1; /* get the normalization factor */
    for (p += k = lvq->cnt; --k >= 0; )
      (--p)->act *= sum;        /* compute rel. neuron activations */
  }                             /* by dividing them by their sum */

  /* --- set result --- */
  if (acts) {                   /* if an activation vector is given */
    for (p += k = lvq->cnt; --k >= 0; )
      acts[k] = (--p)->act;     /* copy the neuron activations */
  }                             /* to the given vector */
  return best;                  /* return index of winner neuron */
}  /* lvq_exec() */

/*--------------------------------------------------------------------*/

DIMID lvq_aggr (LVQNET *lvq, double *vec, double weight)
{                               /* --- aggregate a data vector */
  DIMID  i, k;                  /* prototype index, loop variable */
  QPROT  *p;                    /* to traverse the prototypes */
  double *c, *d;                /* center vector and its change */
  double act, exp;              /* neuron activation and exponent */

  assert(lvq);                  /* check the function arguments */
  i = lvq_exec(lvq, vec, NULL); /* compute degrees of membership */
  lvq->spw += weight;           /* sum weight of processed pattern */
  exp = fabs(lvq->exp);         /* get the adaptation exponent */
  if (exp == 0) exp = 1;        /* no special "winner takes all" */
  for (p = lvq->qps +(k = lvq->cnt); --k >= 0; ) {
    act = (--p)->act;           /* traverse the neurons */
    if (act <= 0) continue;     /* skip unactivated neurons */
    if      (exp == 2) act *= act;
    else if (exp != 1) act = pow(act, exp);
    p->spw += act *= weight;    /* count the training pattern */
    c = p->chg; d = p->dif;     /* get change and difference vector */
    for (i = lvq->dim; --i >= 0; )
      c[i] += act *d[i];        /* agg. the weighted diff. vectors */
    if (!(lvq->type & LVQ_SIZE))/* if there is no size parameter, */
      continue;                 /* skip the size aggregations */
    p->sqr += act *p->d2;       /* sum the squared distances */
    if ((lvq->method & ~LVQ_REPLACE) == LVQ_DIST)
      p->sum += act *sqrt(p->d2);  /* if size based on average dist. */
  }                                /* sum the distances to the center */
  return i;                     /* return index of the winner neuron */
}  /* lvq_aggr() */

/*--------------------------------------------------------------------*/

double lvq_update (LVQNET *lvq)
{                               /* --- update a LVQ network */
  DIMID  i, k;                  /* loop variables */
  QPROT  *p;                    /* to traverse the prototypes */
  double *c, *d;                /* to traverse center and change */
  double max = 0, t;            /* maximum change of a vector */
  double lrcur;                 /* current learning rate */
  double var = 0;               /* adapted uniform variance */

  assert(lvq);                  /* check the function argument */
  for (p = lvq->qps +(k = lvq->cnt); --k >= 0; ) {
    if ((--p)->spw <= 0) {      /* traverse non-empty prototypes */
      p->buf = p->var; continue; }
    t = 1 /p->spw;              /* compute the learning rate */
    lrcur = (lvq->lrcur < t) ? lvq->lrcur : t;
    c = p->ctr; d = p->chg;     /* update the reference vectors */
    for (i = lvq->dim; --i >= 0; ) {
      c[i] += t = lrcur *d[i];  /* update the center coordinates */
      t = fabs(t);              /* determine the maximal change */
      if (t > max) max = t;     /* of a center coordinate and */
      d[i] = 0;                 /* reinitialize the change */
    }
    if (lvq->type & LVQ_SIZE) { /* if to use prototype sizes */
      i = lvq->method & ~LVQ_REPLACE;
      if      (i == LVQ_WEIGHT) {
        t = lvq->spw /(p->spw *(double)lvq->dim);
        p->buf = t *p->var; }   /* compute radius from rel. weight */
      else if (i == LVQ_VAR) {
        t = p->sqr   /(p->spw *(double)lvq->dim);
        p->buf = t; }           /* compute isotropic variance */
      else if (i == LVQ_SQUARE) {
        t = p->sqr /p->spw;     /* compute average squared distance */
        p->buf = t; }
      else {                    /* if (i == LVQ_DIST) */
        t = p->sum /p->spw;     /* compute average distance */
        p->buf = t *t;          /* and square it */
      }
      p->sqr = p->sum = 0;      /* reinit. the sum of (squared) */
    }                           /* distances to the center */
  }
  if (lvq->type & LVQ_SIZE) {   /* if to use prototype sizes */
    if (lvq->szscl < 0) {       /* if to unify the prototype size */
      for (var = 0, p += k = lvq->cnt; --k >= 0; )
        var += sqrt((--p)->buf);/* sum the prototype sizes */
      var *= -lvq->szscl /(double)lvq->cnt;
      var *= var;               /* compute the average size */
    }                           /* and the uniform variance */
    for (p += i = lvq->cnt; --i >= 0; ) {
      --p;                      /* traverse the prototypes */
      if      (lvq->szscl == 0) var = p->var;
      else if (lvq->szscl >  0) var = p->buf *lvq->szscl;
      if (!(lvq->method & LVQ_REPLACE)) {
        lrcur = lvq->lrcur *p->spw;
        if (lrcur > 1) lrcur = 1;
        var = p->var +lrcur *(var -p->var);
      }                         /* compute the new variance */
      p->var = (var < MINVAR) ? MINVAR : var;
    }                           /* set or adapt the variance */
  }                             /* of all prototypes */
  for (p += i = lvq->cnt; --i >= 0; )
    (--p)->spw = 0;             /* reinit. sum of pattern weights */
  lvq->spw = 0;                 /* and the processed pattern weight */
  lvq->steps++;                 /* increment the update step counter */
  if       (lvq->decay < 0)     /* compute the next learning rate */
    lvq->lrcur = lvq->lrate *pow(lvq->steps, lvq->decay);
  else if ((lvq->decay > 0) && (lvq->decay < 1))
    lvq->lrcur = lvq->lrate *pow(lvq->decay, lvq->steps);
  return max;                   /* return the maximal change */
}  /* lvq_update() */

/*--------------------------------------------------------------------*/

void lvq_sort (LVQNET *lvq)
{                               /* --- sort reference vectors */
  DIMID i, s, d, n;             /* loop variables, indices */
  DIMID h;                      /* increment for index */
  QPROT x;                      /* exchange buffer */

  assert(lvq);                  /* check the function argument */
  n = lvq->cnt;                 /* get the number of vectors */
  for (h = n; h > 1; ) {        /* traverse the increments */
    h = (h < 5) ? 1 : ((h *5 -1) /11);
    for (i = h-1; ++i < n; ) {  /* traverse the reference vectors */
      x = lvq->qps[d = i];      /* note the vector to insert and do */
      do {                      /* insertion sort with increment h */
        s = d-h;                /* to find the insertion position */
        if (vec_cmp(lvq->qps[s].ctr, x.ctr, lvq->dim) <= 0)
          break;                /* shift up all vectors that must */
        lvq->qps[d] = lvq->qps[s];      /* follow the current one */
      } while ((d = s) >= h);   /* while another shift is possible */
      lvq->qps[d] = x;          /* store the vector to insert */
    }                           /* at the insertion position */
  }
}  /* lvq_sort() */

/*--------------------------------------------------------------------*/

void lvq_unscale (LVQNET *lvq)
{                               /* --- remove attribute scaling */
  DIMID i;                      /* loop variable */

  assert(lvq);                  /* check the function argument */
  for (i = 0; i < lvq->cnt; i++)/* unscale the center coordinates */
    nst_denorm(lvq->nst, lvq->qps[i].ctr, lvq->qps[i].ctr);
  nst_scale(lvq->nst, -1, 0, 1);
}  /* lvq_unscale() */

/*--------------------------------------------------------------------*/

int lvq_desc (LVQNET *lvq, FILE *file, int mode, int maxlen)
{                               /* --- describe a d.s. network */
  DIMID i, k;                   /* loop variables */
  int   len, l;                 /* loop variables for comments */
  QPROT *p;                     /* to traverse the prototypes */
  char  *indent = "";           /* indentation string */

  assert(lvq && file);          /* check the function arguments */
  len = (maxlen > 0) ? maxlen -2 : 70;

  /* --- print a header (as a comment) --- */
  if (mode & LVQ_TITLE) {       /* if the title flag is set */
    fputs("/*", file); for (l = len; --l >= 0; ) fputc('-', file);
    fputs("\n  vector quantization\n", file);
    for (l = len; --l >= 0; ) fputc('-', file);
    fputs("*/\n", file);        /* print a title header */
  }                             /* (as a comment) */
  if (maxlen <= 0) maxlen = INT_MAX;

  if (!(mode & LVQ_RBFNET))     /* if not part of an RBF network, */
    fputs("lvqnet = {\n",file); /* start the neural network desc. */
  indent = "  ";                /* indent the description itself */

  /* --- print scaling parameters --- */
  nst_desc(lvq->nst, file, indent);

  /* --- print the function description --- */
  fputs(indent,        file);   /* write the indentation */
  fputs("function = ", file);   /* and the function name */
  fputs((lvq->actfn == rf_gauss) ? "gauss" : "cauchy", file);
  fprintf(file,"(%.16g,%.16g);\n", /* write the function parameters */
          lvq->params[0], lvq->params[1]);

  /* --- print the normalization mode --- */
  fputs(indent,        file);   /* write the indentation */
  fputs("normmode = ", file);   /* and the norm. mode indicator */
  fputs(nrmnames[lvq->mode], file);
  if (lvq->wtarf > 0)           /* winner takes all radius factor */
    fprintf(file, ", %.16g", lvq->wtarf);
  fputs(";\n", file);           /* terminate the normalization mode */

  /* --- print network parameters --- */
  fputs(indent,       file);    /* write the indentation and */
  fputs("params = {", file);    /* start the parameter section */
  p = lvq->qps;                 /* traverse the prototypes */
  for (i = 0; i < lvq->cnt; p++, i++) {
    if (i > 0)                  /* start a new output line */
      fprintf(file, ",\n%s          ", indent);
    fprintf(file, "{ [% .16g", p->ctr[0]);
    for (k = 0; ++k < lvq->dim; )
      fprintf(file, ", % .16g", p->ctr[k]);
    fputc(']', file);           /* print the cluster center */
    if (lvq->type & LVQ_SIZE)   /* print the variance (sq. radius) */
      fprintf(file, ",\n%s            [%.16g]", indent, p->var);
    fputs(" }", file);          /* terminate the quant. prototype */
  }
  fputs("};\n", file);          /* terminate the parameter section */

  if (!(mode & LVQ_RBFNET))     /* if not part of an RBF network */
    fputs("};\n", file);        /* terminate the network description */

  /* --- print additional information (as a comment) --- */
  if (mode & LVQ_INFO) {        /* if the add. info. flag is set */
    fputs("\n/*", file); for (l = len; --l >= 0; ) fputc('-', file);
    fprintf(file, "\n  number of inputs : %"DIMID_FMT, lvq->dim);
    fprintf(file, "\n  number of vectors: %"DIMID_FMT, lvq->cnt);
    fputc('\n', file);          /* print additional network info. */
    for (l = len; --l >= 0; ) fputc('-', file);
    fputs("*/\n", file);        /* (as a comment) */
  }
  return ferror(file) ? -1 : 0; /* return the write status */
}  /* lvq_desc() */

/*--------------------------------------------------------------------*/
#ifdef LVQ_PARSE

static int scales (LVQNET *lvq, SCANNER *scan)
{                               /* --- get the input scalings */
  assert(lvq && scan);          /* check the function arguments */
  if (lvq->nst) return 0;       /* check whether scalings exist */
  if ((scn_token(scan) != T_ID) /* check whether 'scales' follows */
  ||  (strcmp(scn_value(scan), "scales") != 0))
    return 0;                   /* if not, abort the function */
  lvq->nst = nst_parse(scan, -1);
  return (lvq->nst) ? 0 : 1;    /* parse the scaling parameters */
}  /* scales() */

/*--------------------------------------------------------------------*/

static int function (LVQNET *lvq, SCANNER *scan)
{                               /* --- parse the function description */
  const char *s;                /* buffer for token value */
  double     p;                 /* buffer for a parameter */

  assert(lvq && scan);          /* check the function arguments */
  if ((scn_token(scan) != T_ID) /* check for a function specification */
  ||  (strcmp(scn_value(scan), "function") != 0))
    return 0;                   /* if there is none, abort */
  SCN_NEXT(scan);               /* consume the "function" keyword */
  SCN_CHAR(scan, '=');          /* consume '=' */
  if (scn_token(scan) != T_ID) SCN_ERROR(scan, E_STREXP, "cauchy");
  s = scn_value(scan);          /* check for a function name */
  if      (strcmp(s, "cauchy") == 0) lvq->actfn = rf_cauchy;
  else if (strcmp(s, "gauss")  == 0) lvq->actfn = rf_gauss;
  else SCN_ERROR(scan, E_STREXP, "cauchy");
  SCN_NEXT(scan);               /* decode and consume it */
  SCN_CHAR(scan, '(');          /* consume '(' */
  if (scn_token(scan) != T_NUM) SCN_ERROR(scan, E_NUMEXP);
  p = strtod(scn_value(scan), NULL);
  if (p <= 0) SCN_ERROR(scan, E_NUMBER);
  lvq->params[0] = p;           /* get the first parameter, */
  SCN_NEXT(scan);               /* i.e., the distance exponent */
  SCN_CHAR(scan, ',');          /* consume the separating ',' */
  if (scn_token(scan) != T_NUM) SCN_ERROR(scan, E_NUMEXP);
  p = strtod(scn_value(scan), NULL);
  if (p < 0) SCN_ERROR(scan, E_NUMBER);
  lvq->params[1] = p;           /* get the second parameter, */
  SCN_NEXT(scan);               /* i.e., the scaling factor */
  SCN_CHAR(scan, ')');          /* consume the closing ')' */
  SCN_CHAR(scan, ';');          /* and the ';' */
  return 0;                     /* return 'ok' */
}  /* function() */

/*--------------------------------------------------------------------*/

static int norm (LVQNET *lvq, SCANNER *scan)
{                               /* --- parse the normalization mode */
  const char *s;                /* buffer for token value */

  assert(lvq && scan);          /* check the function arguments */
  if ((scn_token(scan) != T_ID) /* check for a function specification */
  ||  (strcmp(scn_value(scan), "normmode") != 0))
    return 0;                   /* if there is none, abort */
  SCN_NEXT(scan);               /* consume the "normmode" keyword */
  SCN_CHAR(scan, '=');          /* consume '=' */
  if (scn_token(scan) != T_ID) SCN_ERROR(scan, E_STREXP, "sum1");
  s = scn_value(scan);          /* check for a norm. mode name */
  if      (strcmp(s, "none") == 0) lvq->mode = LVQ_NONE;
  else if (strcmp(s, "sum1") == 0) lvq->mode = LVQ_SUM1;
  else if (strcmp(s, "max1") == 0) lvq->mode = LVQ_MAX1;
  else if (strcmp(s, "hard") == 0) lvq->mode = LVQ_HARD;
  else SCN_ERROR(scan, E_STREXP, "hard");
  SCN_NEXT(scan);               /* decode and consume it */
  if (scn_token(scan) == ',') { /* if a comma follows, */
    SCN_NEXT(scan);             /* consume the comma ',' */
    if (scn_token(scan) != T_NUM) SCN_ERROR(scan, E_NUMEXP);
    lvq->wtarf = strtod(scn_value(scan), NULL);
    SCN_NEXT(scan);             /* get winner takes all radius factor */
  }
  SCN_CHAR(scan, ';');          /* consume the closing ';' */
  return 0;                     /* return 'ok' */
}  /* norm() */

/*--------------------------------------------------------------------*/

static int params (LVQNET *lvq, SCANNER *scan)
{                               /* --- parse a set of prototypes */
  DIMID  i;                     /* loop variable */
  int    err = 0;               /* error status */
  DIMID  vsz = 0;               /* size of prototype vector */
  DIMID  dim = 0;               /* dimension of data space */
  QPROT  *p;                    /* to access the prototypes */
  double *v = NULL;             /* buffer for reallocation */

  assert(lvq && scan);          /* check the function arguments */
  if ((scn_token(scan) != T_ID) /* check for parameters */
  ||  (strcmp(scn_value(scan), "params") != 0))
    SCN_ERROR(scan, E_STREXP, "params");
  SCN_NEXT(scan);               /* consume the "params" keyword */
  SCN_CHAR(scan, '=');          /* consume '=' */
  SCN_CHAR(scan, '{');          /* consume '{' */
  while (1) {                   /* prototype read loop */
    SCN_CHAR(scan, '{');        /* consume '{' */
    if (lvq->cnt >= vsz) {      /* if the prototype vector is full */
      vsz += (vsz > BLKSIZE) ? vsz >> 1 : BLKSIZE;
      p = (QPROT*)realloc(lvq->qps, (size_t)vsz *sizeof(QPROT));
      if (!p) SCN_ERROR(scan, E_NOMEM);
      lvq->qps = p;             /* enlarge the prototype vector */
    }
    SCN_CHAR(scan, '[');        /* consume opening '[' */
    if (lvq->dim <= 0) {        /* if the dimension is yet unknown */
      lvq->dim = 0;             /* initialize the dimension */
      while (1) {               /* coordinate read loop */
        if (scn_token(scan) != T_NUM) SCN_ERROR(scan, E_NUMEXP);
        if (lvq->dim >= dim) {  /* if the coordinate vector is full */
          dim = (dim > BLKSIZE) ? dim >> 1 : BLKSIZE;
          v = (double*)realloc(lvq->vec, (size_t)dim *sizeof(double));
          if (!v) SCN_ERROR(scan, E_NOMEM);
          lvq->vec = v;         /* enlarge the coordinate vector */
        }                       /* and set the new one */
        lvq->vec[lvq->dim++] = strtod(scn_value(scan), NULL);
        SCN_NEXT(scan);         /* get and consume the coordinate */
        if (scn_token(scan) != ',') break;
        SCN_NEXT(scan);         /* if at end of coord. list, abort */
      }                         /* the loop, otherwise consume ',' */
    }
    p = lvq->qps +lvq->cnt++;   /* init. a quantization prototype */
    if (qprot(p, lvq->dim) != 0) SCN_ERROR(scan, E_NOMEM);
    if (dim > 0) {              /* if the dimension was unknown */
      vec_copy(p->ctr, lvq->vec, lvq->dim);
      dim = 0; }                /* copy the prototype center */
    else {                      /* if the dimension was known */
      for (i = 0; i < lvq->dim; i++) {
        if (i > 0) { SCN_CHAR(scan, ','); }  /* consume ',' */
        if (scn_token(scan) != T_NUM) SCN_ERROR(scan, E_NUMEXP);
        p->ctr[i] = strtod(scn_value(scan), NULL);
        SCN_NEXT(scan);         /* traverse the dimensions and */
      }                         /* get and consume the coordinates */
    }                           /* of the center vector */
    SCN_CHAR(scan, ']');        /* consume closing ']' */
    if ((lvq->cnt <= 1)         /* if the prototype type is yet */
    &&  (scn_token(scan) == ','))  /* unknown and there is a comma, */
      lvq->type |= LVQ_SIZE;    /* set the prototype size flag */
    if (lvq->type & LVQ_SIZE) { /* if to use a prototype size */
      SCN_CHAR(scan, ',');      /* consume ',' */
      SCN_CHAR(scan, '[');      /* consume opening '[' */
      if (scn_token(scan) != T_NUM) SCN_ERROR(scan, E_NUMEXP);
      p->var = strtod(scn_value(scan), NULL);
      if (p->var <= 0)             SCN_ERROR(scan, E_NUMBER);
      SCN_NEXT(scan);           /* get and check the squared radius */
      SCN_CHAR(scan, ']');      /* consume closing ']' */
    }
    SCN_CHAR(scan, '}');        /* consume '}' */
    if (scn_token(scan) != ',') break;
    SCN_NEXT(scan);             /* if at end of proto. list, abort */
  }                             /* the loop, otherwise consume ',' */
  SCN_CHAR(scan, '}');          /* consume '}' and ';' */
  SCN_CHAR(scan, ';');          /* (end of the parameters section) */
  return err;                   /* return the error status */
}  /* params() */

/*--------------------------------------------------------------------*/

static LVQNET* parse (SCANNER *scan, DIMID dim)
{                               /* --- parse an LVQ network */
  int    err, e;                /* error status */
  double *v;                    /* to resize the numeric vectors */
  QPROT  *p;                    /* to traverse the prototypes */
  LVQNET *lvq;                  /* created neural network */

  assert(scan);                 /* check the function argument */
  scn_first(scan);              /* get first token */
  lvq = create(dim, 0);         /* create the base structure */
  if (!lvq) return NULL;        /* and initialize some fields */
  err      = scales(lvq, scan); /* read the scaling parameters */
  if (err && (scn_recover(scan, ';', 0, 0, 0) == T_EOF)) {
    lvq_delete(lvq); return NULL; }
  err |= e = function(lvq,scan);/* read the function description */
  if (e   && (scn_recover(scan, ';', 0, 0, 0) == T_EOF)) {
    lvq_delete(lvq); return NULL; }
  err |= e = norm(lvq, scan);   /* read the normalization mode */
  if (e   && (scn_recover(scan, ';', 0, 0, 0) == T_EOF)) {
    lvq_delete(lvq); return NULL; }
  err |= e = params(lvq, scan); /* read the prototype parameters */
  if (e   && (scn_recover(scan, ';', 0, 0, 0) == T_EOF)) {
    lvq_delete(lvq); return NULL; }
  err |=     scales(lvq, scan); /* read the scaling parameters */
  if (err) { lvq_delete(lvq); return NULL; }
  if (!lvq->nst) lvq->nst = nst_create(lvq->dim);
  if (!lvq->nst) { lvq_delete(lvq); return NULL; }
  lvq->qps =                    /* try to shrink the proto. vector */
  p = (QPROT*) realloc(lvq->qps,    (size_t)lvq->cnt *sizeof(QPROT));
  v = (double*)realloc(lvq->vec, 2 *(size_t)lvq->dim *sizeof(double));
  if (!v) { lvq_delete(lvq); return NULL; }
  lvq->vec = v;                 /* create the numeric vectors */
  lvq->buf = v +lvq->dim;       /* and organize them */
  return lvq;                   /* return the created d.s. network */
}  /* parse() */

/*--------------------------------------------------------------------*/

static int header (SCANNER *scan)
{                               /* --- parse header */
  assert(scan);                 /* check the function argument */
  if ((scn_token(scan) != T_ID)
  ||  ((strcmp(scn_value(scan), "lvqnet")   != 0)
  &&   (strcmp(scn_value(scan), "lvq")      != 0)
  &&   (strcmp(scn_value(scan), "clusters") != 0)
  &&   (strcmp(scn_value(scan), "clset")    != 0)))
    SCN_ERROR(scan, E_STREXP, "lvqnet");
  SCN_NEXT(scan);               /* consume 'lvqnet' or 'clset' */
  SCN_CHAR(scan, '=');          /* consume '=' */
  SCN_CHAR(scan, '{');          /* consume '{' */
  return 0;                     /* return 'ok' */
}  /* header() */

/*--------------------------------------------------------------------*/

static int trailer (SCANNER *scan)
{                               /* --- parse trailer */
  assert(scan);                 /* check the function argument */
  SCN_CHAR(scan, '}');          /* consume '}' */
  SCN_CHAR(scan, ';');          /* consume ';' */
  return 0;                     /* return 'ok' */
}  /* trailer() */

/*--------------------------------------------------------------------*/

LVQNET* lvq_parse (SCANNER *scan, DIMID dim, int hdr)
{                               /* --- parse an LVQ network */
  LVQNET *lvq;                  /* created neural network */

  assert(scan);                 /* check the function argument */
  scn_first(scan);              /* get first token */
  if (hdr && (header(scan) != 0))
    return NULL;                /* parse the header if necessary */
  lvq = parse(scan, dim);       /* parse the neural network desc. */
  if (!lvq) return NULL;        /* and then the trailer if necessary */
  if (hdr && (trailer(scan) != 0)) {
    lvq_delete(lvq); return NULL; }
  return lvq;                   /* return the created d.s. network */
}  /* lvq_parse() */

#endif  /* #ifdef LVQ_PARSE */
