/*----------------------------------------------------------------------
  File    : lvqt.c
  Contents: learning vector quantization training
  Author  : Christian Borgelt
  History : 2003.03.11 file created from file cli.c
            2003.05.15 normalization modes added (option -j)
            2003.06.05 multiple parameter retrieval simplified
            2003.06.07 prototype size scaling factor added
            2003.08.15 adapted to new module radfn
            2003.08.16 slight changes in error message output
            2004.02.26 source files mlvqt.c and lvqt.c combined
            2007.01.10 use of attribute maps changed
            2007.02.14 adapted to modified module tabread
            2007.03.16 table and matrix version combined (option -M)
            2007.10.10 evaluation of attribute directions added
            2012.05.15 adapted to modified attset and utility modules
            2013.08.19 adapted to definitions ATTID, VALID, TPLID etc.
            2013.08.29 adapted to new function as_target()
            2014.10.24 changed from LGPL license to MIT license
            2018.09.10 extended functions (attset, attmap) removed
            2019.11.15 adapted to changed interface of mat_shuffle()
----------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <float.h>
#include <time.h>
#include <assert.h>
#ifndef MAT_RDWR
#define MAT_RDWR
#endif
#include "matrix.h"
#ifndef AS_READ
#define AS_READ
#endif
#ifndef AS_PARSE
#define AS_PARSE
#endif
#ifndef AS_DESC
#define AS_DESC
#endif
#include "attset.h"
#include "attmap.h"
#ifndef TAB_READ
#define TAB_READ
#endif
#ifndef TAB_MATRIX
#define TAB_MATRIX
#endif
#include "table.h"
#ifndef LVQ_PARSE
#define LVQ_PARSE
#endif
#include "lvq.h"
#include "random.h"
#include "params.h"
#include "error.h"
#ifdef STORAGE
#include "storage.h"
#endif

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
#define PRGNAME     "lvqt"
#define DESCRIPTION "learning vector quantization training"
#define VERSION     "version 2.3 (2019.11.15)         " \
                    "(c) 2003-2019   Christian Borgelt"

/* --- error codes --- */
/* error codes 0 to -5 defined in attset.h */
#define E_OPTION     (-6)       /* unknown option */
#define E_OPTARG     (-7)       /* missing option argument */
#define E_ARGCNT     (-8)       /* wrong number of arguments */
#define E_PARSE      (-9)       /* parse error */
#define E_TPLCNT    (-10)       /* data file is empty */
#define E_ATTCNT    (-11)       /* no usable attributes found */
#define E_INPCNT    (-12)       /* no usable inputs found */
#define E_UNKTRG    (-13)       /* missing target attribute */
#define E_MULTRG    (-14)       /* multiple target attributes */
#define E_EXCNT     (-15)       /* invalid number of excluded columns */
#define E_MODE      (-16)       /* invalid init. or norm. mode */
#define E_PARAM     (-17)       /* invalid parameter value */
#define E_LRATE     (-18)       /* invalid learning rate */
#define E_METHOD    (-19)       /* invalid variance update method */

#define SEC_SINCE(t)  ((double)(clock()-(t)) /(double)CLOCKS_PER_SEC)

/*--------------------------------------------------------------------*/
#define int         1           /* to check definitions */
#define long        2           /* for certain types */
#define ptrdiff_t   3

#if DIMID!=ATTID
#error "DIMID and ATTID must be the same data type"
#endif

#if DIMID!=TPLID
#error "DIMID and TPLID must be the same data type"
#endif

#undef int                      /* remove preprocessor definitions */
#undef long                     /* needed for the type checking */
#undef ptrdiff_t

/*----------------------------------------------------------------------
  Type Definitions
----------------------------------------------------------------------*/
typedef struct {                /* --- initialization mode info. --- */
  int  code;                    /* code        of initialization mode */
  char *name;                   /* name        of initialization mode */
  char *desc;                   /* description of initialization mode */
} MODEINFO;                     /* (initialization mode information) */

/*----------------------------------------------------------------------
  Constants
----------------------------------------------------------------------*/
static const char *errmsgs[] = {   /* error messages */
  /* E_NONE      0 */  "no error",
  /* E_NOMEM    -1 */  "not enough memory",
  /* E_FOPEN    -2 */  "cannot open file %s",
  /* E_FREAD    -3 */  "read error on file %s",
  /* E_FWRITE   -4 */  "write error on file %s",
  /* E_STDIN    -5 */  "double assignment of standard input",
  /* E_OPTION   -6 */  "unknown option -%c",
  /* E_OPTARG   -7 */  "missing option argument",
  /* E_ARGCNT   -8 */  "wrong number of arguments",
  /* E_PARSE    -9 */  "parse error(s) on file %s",
  /* E_TPLCNT  -10 */  "data file is empty",
  /* E_ATTCNT  -11 */  "no (usable) attributes (need at least 1)",
  /* E_INPCNT  -12 */  "no (usable) inputs (need at least 1)",
  /* E_UNKTRG  -13 */  "unknown target attribute '%s'",
  /* E_MULTRG  -14 */  "multiple target attributes found",
  /* E_EXCNT   -15 */  "invalid number of excluded columns %s",
  /* E_MODE    -16 */  "unknown initialization/normalization mode %s",
  /* E_PARAM   -17 */  "invalid parameter value %g",
  /* E_LRATE   -18 */  "invalid learning rate %g",
  /* E_METHOD  -19 */  "invalid variance update method %s",
  /* E_UNKNOWN -20 */  "unknown error"
};

static const MODEINFO initab[] = { /* table of initialization modes */
  { LVQ_POINTS,  "points",  "randomly chosen points of the dataset"   },
  { LVQ_POINTS,  "pts",                                          NULL },
  { LVQ_POINTS,  "p",                                            NULL },
  { LVQ_UNIFORM, "uniform", "sample from a uniform distribution"      },
  { LVQ_UNIFORM, "unif",                                         NULL },
  { LVQ_UNIFORM, "uni",                                          NULL },
  { LVQ_UNIFORM, "u",                                            NULL },
  { LVQ_CENTER,  "center",  "center of the data space"                },
  { LVQ_CENTER,  "ctr",                                          NULL },
  { LVQ_CENTER,  "c",                                            NULL },
  { LVQ_DIAG,    "diag",    "points on the diagonal of the data space"},
  { LVQ_DIAG,    "d",                                            NULL },
  { LVQ_LATIN,   "latin",   "latin hypercube sampling"                },
  { LVQ_LATIN,   "lhcs",                                         NULL },
  { LVQ_LATIN,   "lhc",                                          NULL },
  { LVQ_LATIN,   "lhs",                                          NULL },
  { LVQ_LATIN,   "l",                                            NULL },
  { -1,          NULL,      /* sentinel */                       NULL },
};

static const MODEINFO nrmtab[] = { /* table of normalization modes */
  { LVQ_NONE,    "none",    "no normalization"                        },
  { LVQ_SUM1,    "sum1",    "normalize to sum 1"                      },
  { LVQ_MAX1,    "max1",    "normalize to maximum 1"                  },
  { LVQ_HARD,    "hard",    "hard assignment"                         },
  { -1,          NULL,      /* sentinel */                       NULL },
};

static const MODEINFO updtab[] = { /* table of update modes */
  { LVQ_DIST,    "dist",    "average distance to center"              },
  { LVQ_DIST,    "dst",                                          NULL },
  { LVQ_DIST,    "d",                                            NULL },
  { LVQ_SQUARE,  "square",  "average squared distance"                },
  { LVQ_SQUARE,  "sqr",                                          NULL },
  { LVQ_SQUARE,  "s",                                            NULL },
  { LVQ_VAR,     "var",     "isotropic variance"                      },
  { LVQ_VAR,     "v",                                            NULL },
  { LVQ_WEIGHT,  "weight",  "weight of assigned patterns"             },
  { LVQ_WEIGHT,  "wgt",                                          NULL },
  { LVQ_WEIGHT,  "w",                                            NULL },
  { -1,           NULL,     /* sentinel */                       NULL },
};

/*----------------------------------------------------------------------
  Global Variables
----------------------------------------------------------------------*/
static CCHAR   *prgname;        /* program name for error messages */
static SCANNER *scan   = NULL;  /* scanner (for domain descriptions) */
static ATTSET  *attset = NULL;  /* attribute set */
static ATTMAP  *attmap = NULL;  /* attribute map */
static TABREAD *tread  = NULL;  /* table reader */
static TABLE   *table  = NULL;  /* table  of training data points */
static MATRIX  *data   = NULL;  /* matrix of training data points */
static LVQNET  *lvq    = NULL;  /* learning vector quantization net */
static FILE    *out    = NULL;  /* network/vector set output file */

/*----------------------------------------------------------------------
  Main Functions
----------------------------------------------------------------------*/

#ifndef NDEBUG
  #undef  CLEANUP               /* clean up memory and close files */
  #define CLEANUP \
  if (lvq)    lvq_delete(lvq);      \
  if (data)   mat_delete(data);     \
  if (table)  tab_delete(table, 0); \
  if (attmap) am_delete(attmap, 0); \
  if (attset) as_delete(attset);    \
  if (tread)  trd_delete(tread, 1); \
  if (scan)   scn_delete(scan,  1); \
  if (out && (out != stdout)) fclose(out);
#endif

GENERROR(error, exit)           /* generic error reporting function */

/*--------------------------------------------------------------------*/

static void help (void)
{                               /* --- print help on init. modes */
  int i;                        /* loop variable */

  fprintf(stderr, "\n");        /* terminate startup message */
  printf("list of initialization modes (option -i#):\n");
  printf("(choice of initial positions of the reference vectors,\n");
  printf("a random offset can be added with the option -o#)\n");
  printf("  name       initialization mode\n");
  for (i = 0; initab[i].name; i++)
    if (initab[i].desc)         /* list of initialization modes */
      printf("  %-9s  %s\n", initab[i].name, initab[i].desc);
  printf("list of activation normalization modes (option -j#)\n");
  printf("  name       activation normalization mode\n");
  for (i = 0; nrmtab[i].name; i++)
    if (nrmtab[i].desc)         /* list of normalization modes */
      printf("  %-9s  %s\n", nrmtab[i].name, nrmtab[i].desc);
  printf("list of variance/radius update methods (option -a#)\n");
  printf("  name       variance update method\n");
  for (i = 0; updtab[i].name; i++)
    if (updtab[i].desc)         /* list of update methods */
      printf("  %-9s  %s\n", updtab[i].name, updtab[i].desc);
  exit(0);                      /* abort the program */
}  /* help() */

/*--------------------------------------------------------------------*/

int main (int argc, char *argv[])
{                               /* --- main function */
  int     i, k = 0;             /* loop variables, buffers */
  char    *s;                   /* to traverse options */
  CCHAR   **optarg = NULL;      /* option argument */
  CCHAR   *fn_inp  = NULL;      /* name of input file */
  CCHAR   *fn_dom  = NULL;      /* name of domain/input vector file */
  CCHAR   *fn_hdr  = NULL;      /* name of table header file */
  CCHAR   *fn_tab  = NULL;      /* name of table file */
  CCHAR   *fn_lvq  = NULL;      /* name of output file */
  CCHAR   *blanks  = NULL;      /* blank   characters */
  CCHAR   *fldseps = NULL;      /* field   separators */
  CCHAR   *recseps = NULL;      /* record  separators */
  CCHAR   *comment = NULL;      /* comment characters */
  CCHAR   *trgname = NULL;      /* name of optional target attribute */
  CCHAR   *ininame = "points";  /* name of initialization mode */
  CCHAR   *nrmname = "hard";    /* name of normalization mode */
  CCHAR   *updname = "dist";    /* name of variance update method */
  int     matinp   = 0;         /* flag for numerical matrix input */
  int     mode     = AS_ATT|AS_NOXATT|AS_NONULL;/* table read mode */
  DIMID   refcnt   = 2;         /* number of reference vectors */
  int     type     = LVQ_CENTER;/* prototype type */
  int     gauss    = 0;         /* flag for Gaussian m.s. function */
  double  params[] = {2, 0};    /* activation function parameters */
  int     nrmmode  = 0;         /* normalization mode */
  double  wtarf    = 0;         /* winner takes all radius factor */
  int     owrite   = 0;         /* overwrite parameters with options */
  int     inimode  = 0;         /* initialization mode */
  double  range    = 0;         /* range for random offsets */
  double  radius   = 1.0;       /* initial radius/std. deviation */
  int     norm     = 1;         /* flag for pattern normalization */
  int     method   = 0;         /* variance update method */
  double  exp      =  0.0;      /* neuron activation exponent */
  double  lrate    =  0.01;     /* initial learning rate */
  double  decay    = -0.1;      /* learning rate decay parameter */
  double  scale    =  1;        /* prototype size scaling factor */
  DIMID   epochs   = 1000, e;   /* number of update epochs */
  DIMID   update   = 1,    u;   /* number of patterns between updates */
  double  trmchg   = 0;         /* maximum change for termination */
  int     shuffle  = 1;         /* whether to shuffle data points */
  int     dmode    = LVQ_TITLE; /* description mode */
  int     maxlen   = 0;         /* maximal output line length */
  DIMID   exclude  = 0;         /* number of target fields to exclude */
  long    seed;                 /* seed for random numbers */
  double  chg, x;               /* (max.) change of a center coord. */
  ATTID   a, c;                 /* number of attributes */
  DIMID   m;                    /* number of inputs/dimensions */
  DIMID   n, r;                 /* number of data points/vectors */
  double  w;                    /* weight of data points/vectors */
  clock_t t;                    /* timer for measurements */

  prgname = argv[0];            /* get program name for error msgs. */

  /* --- print startup/usage message --- */
  if (argc > 1) {               /* if arguments are given */
    fprintf(stderr, "%s - %s\n", argv[0], DESCRIPTION);
    fprintf(stderr, VERSION); } /* print a startup message */
  else {                        /* if no argument given */
    printf("usage: %s [options] [-M|domfile [-d|-h hdrfile]] "
                    "tabfile outfile\n", argv[0]);
    printf("%s\n", DESCRIPTION);
    printf("%s\n", VERSION);
    printf("-!       print a list of init. and norm. modes "
                    "and update methods\n");
    printf("-M       input is a numerical matrix            "
                    "(default: input is a table)\n");
    printf("-c#      number of reference vectors            "
                    "(default: %"DIMID_FMT")\n", refcnt);
    printf("-Z#      use prototype size (initial radius)    "
                    "(default: %g)\n", radius);
    printf("-G       Gaussian activation function           "
                    "(default: Cauchy function)\n");
    printf("-p#:#    activation function parameters         "
                    "(default: %g:%g)\n", params[0], params[1]);
    printf("-j#      activation normalization mode          "
                    "(default: %s)\n", nrmname);
    printf("-y#      winner takes all radius factor         "
                    "(default: %g)\n", wtarf);
    printf("-i#      initialization mode                    "
                    "(default: %s)\n", ininame);
    printf("-o#      random offset range                    "
                    "(default: %g)\n", range);
    printf("-x#      exponent for neuron activations        "
                    "(default: %g)\n", exp);
    printf("-P#      prototype size scaling factor          "
                    "(default: %g)\n", scale);
    printf("-z       replace old size with new size         "
                    "(default: use learning rate)\n");
    printf("-t#:#    initial learning rate and decay        "
                    "(default: %g:%g)\n", lrate, decay);
    printf("-e#      maximum number of update epochs        "
                    "(default: %"DIMID_FMT")\n", epochs);
    printf("-T#      maximum change for termination         "
                    "(default: %g)\n", trmchg);
    printf("-k#      data points between two updates        "
                    "(default: %"DIMID_FMT")\n", update);
    printf("-a#      variance/radius update method          "
                    "(default: %s)\n", updname);
    printf("-O       overwrite with option values           "
                    "(default: parameters read)\n");
    printf("-q       do not normalize input ranges          "
                    "(default: mean=0, var=1)\n");
    printf("-s       do not shuffle data points             "
                    "(default: once per epoch)\n");
    printf("-S#      seed for random numbers                "
                    "(default: time)\n");
    printf("-X#      exclude a target attribute             "
                    "(default: use all atts.)\n");
    printf("-l#      output line length                     "
                    "(default: no limit)\n");
    printf("-b#      blank   characters                     "
                    "(default: \" \\t\\r\")\n");
    printf("-f#      field   separators                     "
                    "(default: \" \\t,\")\n");
    printf("-r#      record  separators                     "
                    "(default: \"\\n\")\n");
    printf("-C#      comment characters                     "
                    "(default: \"#\")\n");
    printf("-n       number of tuple occurrences in last field\n");
    printf("domfile  file containing domain descriptions\n"
           "         (and maybe a set of reference vectors)\n");
    printf("-d       use default header "
                    "(field names = field numbers)\n");
    printf("-h       read table header (field names) from hdrfile\n");
    printf("hdrfile  file containing table header (field names)\n");
    printf("tabfile  table file to read "
                    "(field names in first record)\n");
    printf("outfile  file to write reference vectors to\n");
    printf("infile   file to read  reference vectors from "
                    "(only with -M)\n");
    return 0;                   /* print a usage message */
  }                             /* and abort the program */

  /* --- evaluate arguments --- */
  seed = (long)time(NULL);      /* get a default seed value */
  for (i = 1; i < argc; i++) {  /* traverse arguments */
    s = argv[i];                /* get option argument */
    if (optarg) { *optarg = s; optarg = NULL; continue; }
    if ((*s == '-') && *++s) {  /* -- if argument is an option */
      while (1) {               /* traverse characters */
        switch (*s++) {         /* evaluate option */
          case '!': help();                            break;
          case 'M': matinp  = 1;                       break;
          case 'c': refcnt  = (DIMID)strtol(s, &s, 0); break;
          case 'i': optarg  = &ininame;                break;
          case 'o': range   =      strtod(s, &s);      break;
          case 'G': gauss   = 1;                       break;
          case 'Z': type   |= LVQ_SIZE;
                    radius  = strtod(s, &s);           break;
          case 'p': getdblvec(s, &s, 2, params);       break;
          case 'j': optarg  = &nrmname;                break;
          case 'y': wtarf   =      strtod(s, &s);      break;
          case 'O': owrite  = 1;                       break;
          case 'x': exp     =      strtod(s, &s);      break;
          case 'P': scale   =      strtod(s, &s);      break;
          case 'z': method |= LVQ_REPLACE;             break;
          case 't': getdbls(s, &s, 2, &lrate, &decay); break;
          case 'e': epochs  = (int)strtol(s, &s, 0);   break;
          case 'T': trmchg  =      strtod(s, &s);      break;
          case 'k': update  = (int)strtol(s, &s, 0);   break;
          case 'a': optarg  = &updname;                break;
          case 'q': norm    = 0;                       break;
          case 's': shuffle = 0;                       break;
          case 'S': seed    =      strtol(s, &s, 0);   break;
          case 'X': optarg  = &trgname;                break;
          case 'l': maxlen  = (int)strtol(s, &s, 0);   break;
          case 'b': optarg  = &blanks;                 break;
          case 'f': optarg  = &fldseps;                break;
          case 'r': optarg  = &recseps;                break;
          case 'C': optarg  = &comment;                break;
          case 'n': mode   |= AS_WEIGHT;               break;
          case 'd': mode   |= AS_DFLT;                 break;
          case 'h': optarg  = &fn_hdr;                 break;
          default : error(E_OPTION, *--s);             break;
        }                       /* set option variables */
        if (!*s) break;         /* if at end of string, abort loop */
        if (optarg) { *optarg = s; optarg = NULL; break; }
      } }                       /* get option argument */
    else {                      /* -- if argument is no option */
      switch (k++) {            /* evaluate non-option */
        case  0: fn_dom = s;      break;
        case  1: fn_tab = s;      break;
        case  2: fn_lvq = s;      break;
        default: error(E_ARGCNT); break;
      }                         /* note filenames */
    }
  }
  if (optarg) error(E_OPTARG);  /* check the option argument */
  if (matinp) {                 /* if matrix version */
    if ((k != 2) && (k != 3))   /* check the number */
      error(E_ARGCNT);          /* of arguments */
    fn_inp = fn_lvq;            /* shift the file names */
    fn_lvq = fn_tab;            /* (as there is no domain file) */
    fn_tab = fn_dom; fn_dom = NULL;
    if ((!fn_tab || !*fn_tab) && (!fn_inp || !*fn_inp))
      error(E_STDIN); }         /* stdin must not be used twice */
  else {                        /* if table version */
    if (k != 3) error(E_ARGCNT);/* check the number of arguments */
    if (fn_hdr && (strcmp(fn_hdr, "-") == 0))
      fn_hdr = "";              /* convert "-" to "" */
    i = ( fn_hdr && !*fn_hdr) ? 1 : 0;
    if  (!fn_dom || !*fn_dom) i++;
    if  (!fn_tab || !*fn_tab) i++;
    if (i > 1) error(E_STDIN);  /* check assignments of stdin: */
  }                             /* stdin must not be used twice */
  for (i = 0; initab[i].name; i++)
    if (strcmp(initab[i].name, ininame) == 0) break;
  if (!initab[i].name) error(E_MODE,   ininame);
  inimode = initab[i].code;     /* code the initialization method */
  for (i = 0; nrmtab[i].name; i++)
    if (strcmp(nrmtab[i].name, nrmname) == 0) break;
  if (!nrmtab[i].name) error(E_MODE,   nrmname);
  nrmmode = nrmtab[i].code;     /* code the normalization method */
  for (i = 0; updtab[i].name; i++)
    if (strcmp(updtab[i].name, updname) == 0) break;
  if (!updtab[i].name) error(E_METHOD, updname);
  method |= updtab[i].code;     /* code the update method */
  if (params[0]  <= 0) error(E_PARAM, params[0]);
  if (params[1]  <  0) error(E_PARAM, params[1]);
  if (lrate      <  0) error(E_LRATE, lrate);
  rseed((unsigned)seed);        /* init. the random number generator */
  fputc('\n', stderr);          /* terminate the startup message */

  if (matinp) {                 /* if matrix version */
    /* --- read prototype description --- */
    if (!fn_inp) m = -1;        /* if no input, clear att. counter */
    else {                      /* if an input file is given */
      t    = clock();           /* start the timer */
      scan = scn_create();      /* create a scanner */
      if (!scan) error(E_NOMEM);/* for the input network */
      if (scn_open(scan, NULL, fn_inp) != 0)
        error(E_FOPEN, scn_name(scan));
      fprintf(stderr, "reading %s ... ", scn_name(scan));
      lvq = lvq_parse(scan,0,1);/* parse the network description */
      if (!lvq || !scn_eof(scan, 1))
        error(E_PARSE, scn_name(scan));
      scn_delete(scan, 1);      /* delete the scanner */
      scan = NULL;              /* and clear the variable */
      m      = lvq_dim(lvq);    /* retrieve the number of dimensions */
      refcnt = lvq_cnt(lvq);    /* and the number of ref. vectors */
      fprintf(stderr, "[%"DIMID_FMT" input(s),",  m);
      fprintf(stderr, " %"DIMID_FMT" vector(s)]", refcnt);
      fprintf(stderr, " done [%.2fs].\n", SEC_SINCE(t));
    } }                         /* print a success message */

  else {                        /* if table version */
    /* --- parse domain descriptions --- */
    attset = as_create("domains", att_delete);
    if (!attset) error(E_NOMEM);/* create an attribute set */
    scan = scn_create();        /* create a scanner */
    if (!scan)   error(E_NOMEM);/* for the domain file */
    t = clock();                /* start timer, open input file */
    if (scn_open(scan, NULL, fn_dom) != 0)
      error(E_FOPEN, scn_name(scan));
    fprintf(stderr, "reading %s ... ", scn_name(scan));
    if (as_parse(attset, scan, AT_ALL, 1) != 0)
      error(E_PARSE, scn_name(scan));
    a = as_attcnt(attset);      /* parse domain descriptions */
    if (a < 1) error(E_ATTCNT); /* check for at least one attribute */

    /* --- get optional target attribute --- */
    if (!scn_eof(scan, 0)) trgname = NULL;
    c = as_target(attset, trgname, -1);
    if (c < 0) error(trgname ? E_UNKTRG : E_MULTRG, trgname);
    attmap = am_create(attset, 0, 1.0);
    if (!attmap) error(E_NOMEM);/* create an attribute map */
    if (c < a)                  /* and set the target att. */
      am_target(attmap, c);     /* (prevent that it is used) */
    m = am_inpcnt(attmap);      /* get the number of inputs */
    if (m < 1) error(E_INPCNT); /* check for at least one input */

    /* --- parse network/reference vector set --- */
    if (scn_eof(scan, 0))       /* if there is no input network */
      fprintf(stderr, "[%"ATTID_FMT" attribute(s)]", a);
    else {                      /* if there is an input network, */
      mode |= AS_NOXVAL;        /* prevent nominal domain extensions */
      lvq = lvq_parse(scan,m,1);/* parse the neural network */
      if (!lvq || !scn_eof(scan, 1))
        error(E_PARSE, scn_name(scan));
      fprintf(stderr, "[%"ATTID_FMT" attribute(s),", m);
      fprintf(stderr, " %"DIMID_FMT" vectors(s)]", lvq_cnt(lvq));
    }                           /* print a success message */
    scn_delete(scan, 1);        /* delete the scanner and */
    scan = NULL;                /* clear the scanner variable */
    fprintf(stderr, " done [%.2fs].\n", SEC_SINCE(t));
  }  /* if (matinp) .. else .. */

  if (matinp) {                 /* if matrix version */
    /* --- read training patterns --- */
    tread = trd_create();       /* create a table reader and */
    if (!tread) error(E_NOMEM); /* set the separator characters */
    trd_allchs(tread, recseps, fldseps, blanks, "", comment);
    t = clock();                /* start timer, open input file */
    if (trd_open(tread, NULL, fn_tab) != 0)
      error(E_FOPEN, trd_name(tread));
    fprintf(stderr, "reading %s ... ", trd_name(tread));
    k = mat_readx(&data, tread, -1, m);
    if (k) error(k, TRD_INFO(tread));
    trd_delete(tread, 1);       /* read the training patterns, */
    tread = NULL;               /* then close the input file */
    m = mat_colcnt(data);       /* get the number of variables */
    n = mat_rowcnt(data);       /* and the number of patterns */
    w = (double)n;              /* (which is also the total weight) */
    if (trgname) {              /* if to exclude columns */
      exclude = (DIMID)strtol(trgname, &s, 0);
      if ((s == trgname) || *s || (exclude >= m))
        error(E_EXCNT, trgname);
    }                           /* get number of excluded columns */
    if (exclude > 0)            /* if not to use all columns */
      mat_shrink(data, 0, m -= exclude);
    fprintf(stderr, "[%"DIMID_FMT" input(s),",   m);
    fprintf(stderr, " %"DIMID_FMT" pattern(s)]", n);
    fprintf(stderr, " done [%.2fs].\n", SEC_SINCE(t));
    if (m < 1) error(E_ATTCNT); /* check for at least one attribute */
    if (n < 1) error(E_TPLCNT); /* and   for at least one data point */
  }  /* if (matinp) .. */

  else {                        /* if table version */
    /* --- read table header --- */
    tread = trd_create();       /* create a table reader and */
    if (!tread) error(E_NOMEM); /* set the separator characters */
    trd_allchs(tread, recseps, fldseps, blanks, "", comment);
    if (fn_hdr) {               /* if a header file is given */
      t = clock();              /* start timer, open input file */
      if (trd_open(tread, NULL, fn_hdr) != 0)
        error(E_FOPEN, trd_name(tread));
      fprintf(stderr, "reading %s ... ", trd_name(tread));
      k = as_read(attset, tread, (mode & ~AS_DFLT) | AS_ATT);
      if (k < 0) error(-k, as_errmsg(attset, NULL, 0));
      trd_close(tread);         /* read table header, close file */
      m = as_attcnt(attset);    /* get the number of attributes */
      fprintf(stderr, "[%"ATTID_FMT" attribute(s)]", m);
      fprintf(stderr, " done [%.2fs].\n", SEC_SINCE(t));
      mode &= ~(AS_ATT|AS_DFLT);/* print a success message and */
    }                           /* remove the attribute flag */

    /* --- read data table --- */
    t = clock();                /* start timer, open input file */
    if (trd_open(tread, NULL, fn_tab) != 0)
      error(E_FOPEN, trd_name(tread));
    fprintf(stderr, "reading %s ... ", trd_name(tread));
    table = tab_create("table", attset, tpl_delete);
    if (!table) error(E_NOMEM); /* create a data table */
    k = tab_read(table, tread, mode);
    if (k < 0) error(-k, tab_errmsg(table, NULL, 0));
    trd_delete(tread, 1);       /* read the table body and */
    tread = NULL;               /* delete the table reader */
    a = tab_colcnt(table);      /* get the number of attributes */
    n = tab_tplcnt(table);      /* and the number of data tuples */
    w = tab_tplwgt(table);      /* and print a success message */
    fprintf(stderr, "[%"ATTID_FMT" attribute(s),", a);
    fprintf(stderr, " %"TPLID_FMT, n);
    if (w != (double)n) fprintf(stderr, "/%g", w);
    fprintf(stderr, " tuple(s)] done [%.2fs].\n", SEC_SINCE(t));
    if (n < 1) error(E_TPLCNT); /* check for at least one tuple */

    /* --- convert into matrix --- */
    data = tab_matrix(table, attmap, AM_INPUTS, 1);
    if (!data) error(E_NOMEM);  /* convert the read table */
    table = NULL;               /* into a (data) matrix */
  }  /* if (matinp) .. else .. */

  /* --- reduce data --- */
  if (update <= 0) {            /* if to update once each epoch */
    fprintf(stderr, "reducing %s ... ", (matinp) ? "matrix" : "table");
    n = mat_reduce(data);       /* reduce matrix for speed up */
    fprintf(stderr, "[%"DIMID_FMT, n);
    if (w != (double)n) fprintf(stderr, "/%g", w);
    fprintf(stderr, (matinp) ? " vector" : " tuple");
    fprintf(stderr, "(s)] done [%.2fs].\n", SEC_SINCE(t));
  }                             /* print a log message */

  /* --- create network/reference vector set --- */
  if (lvq)                      /* if an input network is given */
    refcnt = lvq_cnt(lvq);      /* get the number of ref. vectors */
  else {                        /* if no input network is given */
    t = clock();                /* start the timer, print message */
    fprintf(stderr, "creating network ... ");
    lvq = lvq_create(mat_colcnt(data), refcnt);
    if (!lvq) error(E_NOMEM);   /* create a new network */
    n = mat_rowcnt(data);       /* get the number of data points */
    for (r = 0; r < n; r++)     /* determine ranges of values */
      lvq_reg(lvq, mat_row(data, r), 1);
    if (norm)                   /* if to normalize the data, */
      lvq_reg(lvq, NULL, 0);    /* compute the scaling factors */
    if (inimode < LVQ_POINTS)   /* if not to use the training data */
      lvq_init(lvq, inimode, range, drand, NULL);
    else {                      /* if to select random data points, */
      mat_shuffle(data, drand, 0, -1);      /* shuffle the patterns */
      for (r = refcnt; --r >= 0; ) {
        lvq_init(lvq, inimode, range, drand,
                 mat_row(data, (DIMID)r % n));
      }                         /* use the first refcnt patterns */
    }                           /* as the initial reference vectors */
    owrite = -1;                /* set the overwrite flag */
    fprintf(stderr, "[%"DIMID_FMT" vector(s)]", refcnt);
    fprintf(stderr, " done [%.2fs].\n", SEC_SINCE(t));
  }                         /* print a success message */

  /* --- train neural network --- */
  t = clock();                  /* start timer, print message */
  fprintf(stderr, "training neural network ... ");
  if (owrite) {                 /* if to overwrite the parameters */
    lvq_type(lvq, type, radius);/* set the prototype type */
    lvq_actfn(lvq, (gauss) ? rf_gauss : rf_cauchy, params);
  }                             /* set the activation function */
  lvq_norm  (lvq, nrmmode, wtarf);   /* set the normalization mode, */
  lvq_lrate (lvq, lrate, decay);/* the learning rate and its decay, */
  lvq_exp   (lvq, exp);         /* the adaptation exponent, */
  lvq_scale (lvq, scale);       /* the size scaling factor, */
  lvq_method(lvq, method);      /* and the update method */
  if (update <= 0) shuffle = 0; /* suppress unnecessary shuffling */
  u = update;                   /* initialize the tuple counter */
  for (e = 0; e < epochs; e++){ /* compute a maximum number of epochs */
    if ((e & 0x00ff) == 0)      /* print the current number of epochs */
      fprintf(stderr, "%8"DIMID_FMT"\b\b\b\b\b\b\b\b", e);
    chg = 0;                    /* clear the maximal change */
    if (shuffle || (e == 0))    /* shuffle the training patterns */
      mat_shuffle(data, drand, 0, -1);  /* (at least in epoch 0) */
    for (r = mat_rowcnt(data); --r >= 0; ) {
      lvq_aggr(lvq, mat_row(data, r), 1);
      if ((update > 0) && (--u <= 0)) {
        u = update;             /* update the prototypes */
        x = lvq_update(lvq);    /* every 'update' patterns */
        if (x > chg) chg = x;   /* determine the maximum change */
      }                         /* of a center coordinate */
    }                           /* if (matinp) .. else .. */
    if (update <= 0)            /* if no number of patterns is given, */
      chg = lvq_update(lvq);    /* update once in each epoch */
    if (chg <= trmchg) break;   /* check the termination criterion */
  }                             /* write a success message */
  fprintf(stderr, "[%"DIMID_FMT" epoch(s)]", e);
  fprintf(stderr, " done [%.2fs].\n", SEC_SINCE(t));

  /* --- write neural network --- */
  t = clock();                  /* start timer, open output file */
  if (fn_lvq && (strcmp(fn_lvq, "-") == 0)) fn_lvq = "";
  if (fn_lvq && *fn_lvq) { out = fopen(fn_lvq, "w"); }
  else                   { out = stdout; fn_lvq = "<stdout>"; }
  fprintf(stderr, "writing %s ... ", fn_lvq);
  if (!out) error(E_FOPEN, fn_lvq);
  if (!matinp) {                /* if table version */
    if (as_desc(attset, out, AS_TITLE|AS_MARKED|AS_IVALS, maxlen) != 0)
      error(E_FWRITE, fn_lvq);  /* describe attribute domains */
    fprintf(out, "\n");         /* leave one line empty */
  }
  lvq_sort(lvq);                /* sort the reference vectors */
  if (lvq_desc(lvq, out, dmode, maxlen) != 0)
    error(E_FWRITE, fn_lvq);    /* describe the reference vectors */
  if (out != stdout) {          /* if not written to standard output, */
    i = fclose(out); out = NULL;/* close the output file */
    if (i != 0) error(E_FWRITE, fn_lvq);
  }                             /* print a log message */
  fprintf(stderr, "[%"DIMID_FMT" vector(s)]", refcnt);
  fprintf(stderr, " done [%.2fs].\n", SEC_SINCE(t));

  /* --- clean up --- */
  CLEANUP;                      /* clean up memory and close files */
  SHOWMEM;                      /* show (final) memory usage */
  return 0;                     /* return 'ok' */
}  /* main() */
